using System;
using System.Collections.Generic;
using System.Text;

namespace Sketcher {

    /*
     * RenderingBehaviors are special-case rendering actions that would fall outside
     *  the standard rendering loop (ie drawing to overlay planes, etc)
     */
    abstract class RenderingBehavior : Behavior {
        public RenderingBehavior(Object target)
            : base(target) {
            Connect();
        }

        //! connect this behavior to the world
        override public void Connect() {
            RenderingBehaviorManager.Get.Register(this);
        }

        //! disconnect this behavior from the world
        override public void Disconnect() {
            RenderingBehaviorManager.Get.DeRegister(this);
        }


        //! Called when Overlay background has been painted. Return true to indicate that overlay background
        //!  has been updated and should be stored (ie if you appended parts of a stroke to it, etc)
        virtual public bool OnUpdateOverlayBackground() {
            return false;
        }
        //! OnOverlayUpdateBackground only called if this returns true
        virtual public bool UpdateOverlayFlag {
            get { return false; }
        }


        //! called when Overlay background has been painted, but before flipping buffers. Return
        //!  true to indicate that something was drawn (ie that overlay is dirty)
        virtual public bool OnOverlayRender() {
            return false;
        }
        //! OnOverlayRender only called if this returns true
        virtual public bool OverlayRenderFlag {
            get { return false; }
        }

    }




    /*
     * "versions" of class with different types of target
     */

    abstract class ElementRenderingBehavior : RenderingBehavior {
        public ElementRenderingBehavior(Element target)
            : base(target) {
        }

        public Element TargetElement {
            get { return Target as Element; }
        }

        //! behaviors are only considered for enabled items
        override public bool Enabled {
            get { return Element.VisibleForRendering(TargetElement.Visibility); }
        }
    }


    abstract class ItemRenderingBehavior : ElementRenderingBehavior {
        public ItemRenderingBehavior(Item target)
            : base(target) {
        }

        public Item TargetItem {
            get { return Target as Item; }
        }
    }



    abstract class WidgetRenderingBehavior : ElementRenderingBehavior {
        public WidgetRenderingBehavior(Widget target)
            : base(target) {
        }

        public Widget TargetWidget {
            get { return Target as Widget; }
        }

    }


}
