using System;
using System.Collections.Generic;
using System.Text;

using Tao.OpenGl;

namespace Sketcher
{
    //! an Item is a UI Element that must be placed in a Container and will be
    //! positioned by a Layout. (Containers are also Items)
    abstract class Item : Element
    {
        Transform m_itemTransform;


        public Item() {

            m_itemTransform = new Transform( );
            m_transform.PushTransform( m_itemTransform );

            m_fgColor = ColorRGBA.BLACK;
            m_bgColor = ColorRGBA.WHITE;
            m_fOpacity = 1.0f;
            m_Visibility = VisibilityState.Visible;
        }


        override public Element Parent {
            set {
                m_parent = value;
                m_transform.Clear();
                if (m_parent != null) {
                    Container c = value as Container;
                    if (c == null)
                        throw new Exception( "[Item] parent of Item must be a Container" );
                    m_transform.PushTransform( c.ChildAreaTransformation );
                }
                m_transform.PushTransform( m_itemTransform );
            }
        }

        public Container ParentContainer {
            get { return Parent as Container; }
        }

        public Transform ItemTransform {
            get { return m_itemTransform; }
            set { m_itemTransform = value; }
        }

        /*
         * Item Interface
         */
#region Item Interface

        virtual public IRegion Geometry {
            get { return Bounds; }
        }

        virtual public void Resize(Box2f newBounds) {
            Resize(newBounds, null);
        }
        virtual public void Resize(Box2f newBounds, Transform relativeTo) {
            m_size.Width = newBounds.Width;
            m_size.Height = newBounds.Height;

            // set relative bounds...
            if (relativeTo != null) {
                newBounds = relativeTo.Apply(newBounds, true);
            }

            m_itemTransform.Clear();
            m_itemTransform.PushTransform( newBounds.Min );

            NotifyItemBoundsChanged();
            NotifyElementChanged();
        }


        virtual public bool CanResizeWidth {
            get { return true; }
        }
        virtual public bool CanResizeHeight {
            get { return true; }
        }


#endregion


        // notification handler for item changes
        public delegate void ItemBoundsChangedHandler(Item item);
        public event ItemBoundsChangedHandler ItemBoundsChangedEvent;

        protected void NotifyItemBoundsChanged() {
            if (ItemBoundsChangedEvent != null)
                ItemBoundsChangedEvent(this);
        } 




#region Renderable Members

        override public VisibilityState Visibility {
            get {
                if ( Parent != null )
                    return Element.DetermineVisibility(m_Visibility, Parent.Visibility);
                return m_Visibility;
            }
        }

        //! overrides must implement this!
        override abstract public void Render();

        //! can optionally implement this
        override public void RenderBackground() {}


#endregion
    }
}
