using System;
using System.Collections.Generic;
using System.Text;
using Tao.OpenGl;
using Tao.Platform.Windows;

namespace Sketcher
{
    /*
     *  This class basically implements various methods to save and restore a rendering.
     *  (either render-to-texture or direct framebuffer storage). 
     */

    class RenderCache
    {
        public interface RenderCacheTarget
        {
            int Width { get; }
            int Height { get; }
            void RenderCacheableContent( );
        }


        public enum CacheModes {
            Unknown,
            ARBBufferRegion,
            GLCopyPixels,
            FrameBufferObject
        };
        CacheModes m_eMode;

        RenderCacheTarget m_canvas;

        public RenderCache( RenderCacheTarget target ) {
            m_eMode = CacheModes.Unknown;
            m_canvas = target;
        }


        public RenderCacheTarget Target {
            get { return m_canvas; }
        }


        public CacheModes RenderMode {
            get {
                if (m_eMode == CacheModes.Unknown)
                    Initialize();
                return m_eMode;
            }
        }

        //! can only be called once GL context is set up
        static bool s_glFunctionsReloaded = false;
        static bool s_wglFunctionsReloaded = false;
        public void Initialize( ) {
            if (!s_glFunctionsReloaded) {
                Gl.ReloadFunctions();
                s_glFunctionsReloaded = true;
                string sVersion = Gl.glGetString( Gl.GL_VERSION );
                string[] sExtensions = Gl.glGetString( Gl.GL_EXTENSIONS ).Split( ' ' );
            }
            bool bSupportsExtFrameBufferObject = Gl.IsExtensionSupported( "GL_EXT_framebuffer_object" );

            if (!s_wglFunctionsReloaded) {
                Wgl.ReloadFunctions();
                s_wglFunctionsReloaded = true;
                string[] sWglExtesions = Wgl.wglGetExtensionsStringEXT().Split( ' ' );
            }

            bool bSupportsARBBufferRegion = Wgl.IsExtensionSupported( "WGL_ARB_buffer_region" );
            bool bSupportsPBuffer = Wgl.IsExtensionSupported( "WGL_ARB_pbuffer" );

            //bSupportsExtFrameBufferObject = false;

            if (bSupportsExtFrameBufferObject && InitializeFrameBufferObject()) {
                m_eMode = CacheModes.FrameBufferObject;
                System.Diagnostics.Debug.WriteLine( "[GLOverlay] using Frame Buffer Object" );
            } else if (bSupportsARBBufferRegion && InitializeARBBufferRegion()) {
                m_eMode = CacheModes.ARBBufferRegion;
                System.Diagnostics.Debug.WriteLine( "[GLOverlay] using ARB Buffer Region" );
            } else if (InitializeGLCopyPixels()) {
                m_eMode = CacheModes.GLCopyPixels;
                System.Diagnostics.Debug.WriteLine( "[GLOverlay] using glCopyPixels" );
            } else {
                m_eMode = CacheModes.Unknown;
            }
        }


        //! make sure cache is valid - IE has stored render of scene
        public bool Validate() {
            switch ( RenderMode ) {
                case CacheModes.GLCopyPixels:
                    return StoreGLCopyPixels();

                case CacheModes.ARBBufferRegion:
                    // [TODO] don't always need to do a full paint here - need to set a flag
                    //   somewhere to determine if backbuffer has already been painted...
                    return StoreARBBufferRegion(true);

                case CacheModes.FrameBufferObject:
                    return StoreFrameBufferObject();
            }
            return false;
        }

        //! notify cache that it needs to update itself
        public void Invalidate() {
            m_bARBBufferRegionValid = false;
            m_bFBOValid = false;
            m_bGLCopyPixelsValid = false;
        }

        //! start manual update of cache (IE draw to cache directly, w/o invalidation)
        public bool BeginUpdate() {
            switch ( RenderMode ) {
                case CacheModes.FrameBufferObject:
                    return BeginUpdateFrameBufferObject();
                case CacheModes.ARBBufferRegion:
                    return BeginUpdateARBBufferRegion();
                case CacheModes.GLCopyPixels:
                    return BeginUpdateGLCopyPixels();
            }
            return false;
        }

        //! notify cache that we are done updating (save result of update)
        public bool DoneUpdate() {
            switch (RenderMode) {
                case CacheModes.FrameBufferObject:
                    return DoneUpdateFrameBufferObject();
                case CacheModes.ARBBufferRegion:
                    return DoneUpdateARBBufferRegion();
                case CacheModes.GLCopyPixels:
                    return DoneUpdateGLCopyPixels();
            }
            return false;
        }

        //! draw the cache into the current framebuffer
        public bool RestoreFromCache() {
            switch (RenderMode) {
                case CacheModes.ARBBufferRegion:
                    return RestoreARBBufferRegion();

                case CacheModes.FrameBufferObject:
                    return RestoreFrameBufferObject();

                case CacheModes.GLCopyPixels:
                    return RestoreGLCopyPixels();

                default:
                    System.Diagnostics.Debugger.Break();
                    return false;                    
            }
        }




        /*
         * FrameBufferObject overlay cache implementation
         *    - currently most efficient approach (as of Feb 2009)
         *    - completely unsupported on older cards, or on Intel chips (apparenlty)
         */
        uint m_FBO_Overlay_id = 0;
        uint m_FBO_RenderBuffer_Overlay_id = 0;
        uint m_FBO_Texture_id = 0;
        bool m_bFBOValid = false;
        int m_nFBOSize = 0;

        bool InitializeFrameBufferObject() {
            if ( m_FBO_Overlay_id == 0 ) {

                // clear errors
                GLCanvas.CheckGLErrors();

                int nMaxTexSize, nMaxRenderBufSize;
                Gl.glGetIntegerv(Gl.GL_MAX_TEXTURE_SIZE, out nMaxTexSize);
                Gl.glGetIntegerv(Gl.GL_MAX_RENDERBUFFER_SIZE_EXT, out nMaxRenderBufSize);

                // determine power-of-two FBO size for current viewport
                int nMaxSize = Math.Max( Target.Width, Target.Height );
                m_nFBOSize = 1;
                while (m_nFBOSize < nMaxSize)
                    m_nFBOSize *= 2;
                if (m_nFBOSize > nMaxTexSize || m_nFBOSize > nMaxRenderBufSize) {
                    System.Windows.Forms.MessageBox.Show("Cannot allocate FrameBufferObject - necessary size " + m_nFBOSize + " is larger than max texture size " + nMaxTexSize + " or max renderbuffer size " + nMaxRenderBufSize);
                    return false;
                }

                // create FBO
                Gl.glGenFramebuffersEXT(1, out m_FBO_Overlay_id);

                // create texture for FBO
                Gl.glGenTextures(1, out m_FBO_Texture_id);
                Gl.glBindTexture(Gl.GL_TEXTURE_2D, m_FBO_Texture_id);
                Gl.glTexImage2D(Gl.GL_TEXTURE_2D, 0, Gl.GL_RGBA8, m_nFBOSize, m_nFBOSize,
                    0, Gl.GL_RGBA, Gl.GL_UNSIGNED_BYTE, IntPtr.Zero);

                // [RMS] this generates mipmaps, but we don't need them for overlay...
                //Gl.glGenerateMipmapEXT(Gl.GL_TEXTURE_2D);
                //Gl.glTexParameterf(Gl.GL_TEXTURE_2D, Gl.GL_TEXTURE_MIN_FILTER, Gl.GL_LINEAR_MIPMAP_LINEAR);

                // [RMS] have to set these tex parameters or check below will fail...
                //   (actually it seems like only MIN_FILTER and MAG_FILTER must be set...)
                Gl.glTexParameterf(Gl.GL_TEXTURE_2D, Gl.GL_TEXTURE_MIN_FILTER, Gl.GL_LINEAR);
                Gl.glTexParameterf(Gl.GL_TEXTURE_2D, Gl.GL_TEXTURE_MAG_FILTER, Gl.GL_LINEAR);
                Gl.glTexParameterf(Gl.GL_TEXTURE_2D, Gl.GL_TEXTURE_WRAP_S, Gl.GL_CLAMP_TO_EDGE);
                Gl.glTexParameterf(Gl.GL_TEXTURE_2D, Gl.GL_TEXTURE_WRAP_T, Gl.GL_CLAMP_TO_EDGE);

                if (GLCanvas.CheckGLErrors())
                    System.Diagnostics.Debugger.Break();

                // create depth renderbuffer
                Gl.glGenRenderbuffersEXT(1, out m_FBO_RenderBuffer_Overlay_id);
                Gl.glBindRenderbufferEXT(Gl.GL_RENDERBUFFER_EXT, m_FBO_RenderBuffer_Overlay_id);
                Gl.glRenderbufferStorageEXT(Gl.GL_RENDERBUFFER_EXT, Gl.GL_DEPTH_COMPONENT, m_nFBOSize, m_nFBOSize);

                // create and bind FBO
                Gl.glBindFramebufferEXT(Gl.GL_FRAMEBUFFER_EXT, m_FBO_Overlay_id);
                
                // attach texture and depth renderbuffer to FBO
                Gl.glFramebufferTexture2DEXT(Gl.GL_FRAMEBUFFER_EXT, Gl.GL_COLOR_ATTACHMENT0_EXT,
                    Gl.GL_TEXTURE_2D, m_FBO_Texture_id, 0);
                Gl.glFramebufferRenderbufferEXT(Gl.GL_FRAMEBUFFER_EXT, Gl.GL_DEPTH_ATTACHMENT_EXT,
                    Gl.GL_RENDERBUFFER_EXT, m_FBO_RenderBuffer_Overlay_id);

                // error check
                int nStatus = Gl.glCheckFramebufferStatusEXT(Gl.GL_FRAMEBUFFER_EXT);
                if (nStatus != Gl.GL_FRAMEBUFFER_COMPLETE_EXT) {
                    System.Diagnostics.Debugger.Break();
                    return false;
                }

                // disconnect FBO from window
                Gl.glBindFramebufferEXT(Gl.GL_FRAMEBUFFER_EXT, 0);

                m_bFBOValid = false;
            }
            return m_FBO_Overlay_id != 0;
        }

        bool StoreFrameBufferObject() {
            // [TODO] need to be able to re-generate FBO if texture is too small!
            if (Target.Width > m_nFBOSize || Target.Height > m_nFBOSize)
                System.Diagnostics.Debugger.Break();

            if (!m_bFBOValid && m_FBO_Overlay_id != 0) {

                // render to FBO
                Gl.glBindFramebufferEXT(Gl.GL_FRAMEBUFFER_EXT, m_FBO_Overlay_id);

                // want to draw with same viewport as main window
                Gl.glPushAttrib(Gl.GL_VIEWPORT_BIT);
                Gl.glViewport( 0, 0, Target.Width, Target.Height );

                Target.RenderCacheableContent();

                // disconnect
                Gl.glPopAttrib();
                Gl.glBindFramebufferEXT(Gl.GL_FRAMEBUFFER_EXT, 0);

                m_bFBOValid = true;
            }
            return m_bFBOValid;
        }


        bool BeginUpdateFrameBufferObject() {
            // [TODO] need to be able to re-generate FBO if texture is too small!
            if (Target.Width > m_nFBOSize || Target.Height > m_nFBOSize)
                System.Diagnostics.Debugger.Break();

            if ( m_FBO_Overlay_id != 0 ) {
                // render to FBO
                Gl.glBindFramebufferEXT(Gl.GL_FRAMEBUFFER_EXT, m_FBO_Overlay_id);

                // want to draw with same viewport as main window
                Gl.glPushAttrib(Gl.GL_VIEWPORT_BIT);
                Gl.glViewport( 0, 0, Target.Width, Target.Height );

                return true;
            }
            return false;
        }
        bool DoneUpdateFrameBufferObject() {
            if ( m_FBO_Overlay_id != 0 ) {
                // pop viewport setting
                Gl.glPopAttrib();

                // disconnect from FBO
                Gl.glBindFramebufferEXT(Gl.GL_FRAMEBUFFER_EXT, 0);
                return true;
            }
            return false;
        }

        bool RestoreFrameBufferObject() {
            if (m_bFBOValid) {
                Gl.glPushAttrib(Gl.GL_ENABLE_BIT);
                Gl.glEnable(Gl.GL_TEXTURE_2D);

                Gl.glBindTexture(Gl.GL_TEXTURE_2D, m_FBO_Texture_id);
                Gl.glColor4f(1, 1, 1, 1);

                float fX = (float)Target.Width / (float)m_nFBOSize;
                float fY = (float)Target.Height / (float)m_nFBOSize;
                Gl.glBegin(Gl.GL_QUADS);
                Gl.glTexCoord2f(0, 0);
                Gl.glVertex2f(0, 0);
                Gl.glTexCoord2f(fX, 0);
                Gl.glVertex2f( Target.Width, 0 );
                Gl.glTexCoord2f(fX, fY);
                Gl.glVertex2f( Target.Width, Target.Height );
                Gl.glTexCoord2f(0, fY);
                Gl.glVertex2f( 0, Target.Height );
                Gl.glEnd();

                Gl.glBindTexture(Gl.GL_TEXTURE_2D, 0);

                Gl.glPopAttrib();

                return true;
            }
            return false;
        }



        /*
         * ARBBufferRegion overlay cache implementation
         *   - [TODO] check if this actually works
         */
        IntPtr m_ARBBufferRegionHandle = IntPtr.Zero;
        bool m_bARBBufferRegionValid = false;

        bool InitializeARBBufferRegion() {
            if (m_ARBBufferRegionHandle == IntPtr.Zero) {
                m_ARBBufferRegionHandle = Wgl.wglCreateBufferRegionARB(
                    Wgl.wglGetCurrentDC(), 0, Wgl.WGL_BACK_COLOR_BUFFER_BIT_ARB);
                if (GLCanvas.CheckGLErrors())
                    System.Diagnostics.Debugger.Break();
                m_bARBBufferRegionValid = false;
            }
            return m_ARBBufferRegionHandle != IntPtr.Zero;
        }

        bool StoreARBBufferRegion(bool bTargetMustRepaint) {
            if (!m_bARBBufferRegionValid && m_ARBBufferRegionHandle != IntPtr.Zero) {

                if ( bTargetMustRepaint )
                    Target.RenderCacheableContent();

                Wgl.wglSaveBufferRegionARB( m_ARBBufferRegionHandle, 
                    0, 0, Target.Width, Target.Height );
                m_bARBBufferRegionValid = true;
            } 
            return m_bARBBufferRegionValid;
        }

        bool RestoreARBBufferRegion() {
            if (m_bARBBufferRegionValid) {
                Wgl.wglRestoreBufferRegionARB( m_ARBBufferRegionHandle, 
                    0, 0, Target.Width, Target.Height, 0, 0 );
                return true;
            }
            return false;
        }

        bool BeginUpdateARBBufferRegion( ) {
            return RestoreARBBufferRegion();
        }
        bool DoneUpdateARBBufferRegion( ) {
            return StoreARBBufferRegion(false);
        }



        /*
         * GLCopyPixels overlay cache implementation
         *    - transfers to and from graphics card for each update - very slow!!
         *    - doesn't use any graphics card memory
         */
        byte[] m_pixelBuf = null;
        bool m_bGLCopyPixelsValid = false;

        bool InitializeGLCopyPixels( ) {
            m_pixelBuf = new byte[Target.Width*Target.Height*4];
            m_bGLCopyPixelsValid = false;
            return true;
        }

        bool StoreGLCopyPixels( ) {
            if (m_pixelBuf.Length < Target.Width * Target.Height * 4) {
                m_pixelBuf = new byte[Target.Width*Target.Height*4];
            }
            if (m_pixelBuf != null) {
                unsafe {
                    fixed (byte* pArray = m_pixelBuf) {
                        IntPtr intPtr = new IntPtr( (void*)pArray );
                        Gl.glReadPixels( 0, 0, Target.Width, Target.Height, Gl.GL_RGBA, Gl.GL_UNSIGNED_BYTE, intPtr );
                        m_bGLCopyPixelsValid = true;
                    }
                }
                return true;
            }
            return false;
        }


        bool RestoreGLCopyPixels( ) {
            if (m_bGLCopyPixelsValid) {
                unsafe {
                    fixed (byte* pArray = m_pixelBuf) {
                        IntPtr intPtr = new IntPtr( (void*)pArray );
                        Gl.glRasterPos2i( 0, 0 );
                        Gl.glDrawPixels( Target.Width, Target.Height, Gl.GL_RGBA, Gl.GL_UNSIGNED_BYTE, intPtr );
                    }
                }
                return true;
            }
            return false;
        }


        bool BeginUpdateGLCopyPixels( ) {
            return RestoreGLCopyPixels();
        }
        bool DoneUpdateGLCopyPixels( ) {
            return StoreGLCopyPixels();
        }




    }
}
