using System;
using System.Collections.Generic;
using System.Text;
using Tao.OpenGl;

namespace Sketcher
{
    class GLTexture
    {
        Image m_image;

        int m_nGLTextureID;         // current texture ID (0 if invalid)
        int m_nDeleteGLTextureID;   // texture ID to free when we are sure we have context

	    int m_nGLTextureFormat;	    // GL_RGB / RGBA / etc
	    int m_nGLTextureType;		// GL_TEXURE_1D / 2D / 3D
	    int m_nGLTextureEnvMode;	// GL_MODULATE / DECAL / REPLACE / etc
	    int m_nGLTextureFilter;	    // GL_NEAREST / GL_LINEAR
	    int m_nGLTextureWrapMode;	// GL_REPEAT / GL_CLAMP

        int m_nTexWidth, m_nTexHeight;   // power-of-two texure size
        float m_fScaleX, m_fScaleY;      // scaling factor for non-power-of-two textures


        public GLTexture( Image img ) {
            m_nDeleteGLTextureID = m_nGLTextureID = 0;
            m_nGLTextureType = Gl.GL_TEXTURE_2D;
//            m_nGLTextureEnvMode = Gl.GL_MODULATE;
            m_nGLTextureEnvMode = Gl.GL_REPLACE;
            m_nGLTextureFilter = Gl.GL_LINEAR;
            m_nGLTextureWrapMode = Gl.GL_CLAMP;
            m_nTexWidth = m_nTexHeight = 0;
            m_fScaleX = m_fScaleY = 1;

            m_image = img;
            OnImageChanged();
        }

        public Image TexImage {
            get { return m_image; }
            set { m_image = value; OnImageChanged(); }
        }

        public int TextureID {
            get { return GetTextureID(); }
        }

        //! can be GL_REPLACE, GL_MODULATE, GL_DECAL, (??)
        public int TexEnvMode {
            get { return m_nGLTextureEnvMode; }
            set { m_nGLTextureEnvMode = value; }
        }

        //! can be GL_NEAREST or GL_LINEAR
        public int TextureFilter {
            get { return m_nGLTextureFilter; }
            set { m_nGLTextureFilter = value; }
        }

        public bool Enable() {
            int nTextureID = GetTextureID();
            if (nTextureID != 0) {
                Gl.glEnable(m_nGLTextureType);
                Gl.glBindTexture(m_nGLTextureType, nTextureID);

                Gl.glTexEnvi(Gl.GL_TEXTURE_ENV, Gl.GL_TEXTURE_ENV_MODE, m_nGLTextureEnvMode);

                Gl.glTexParameteri( m_nGLTextureType, Gl.GL_TEXTURE_MIN_FILTER, m_nGLTextureFilter );
                Gl.glTexParameteri( m_nGLTextureType, Gl.GL_TEXTURE_MAG_FILTER, m_nGLTextureFilter );

		        // apply texture scaling to get image into the right position...
		        Gl.glPushAttrib( Gl.GL_TRANSFORM_BIT );
		        Gl.glMatrixMode( Gl.GL_TEXTURE);
		        Gl.glPushMatrix();
		        Gl.glLoadIdentity();
		        Gl.glTranslatef( 0.5f, 0.5f, 0.0f );
		        Gl.glScalef( m_fScaleX, m_fScaleX, 0.0f );
		        Gl.glTranslatef( -0.5f, -0.5f, 0.0f );
		        Gl.glPopAttrib();

                return true;
            } else
                return false;
        }

        public void Disable() {
	        Gl.glBindTexture( m_nGLTextureType, 0 );
	        Gl.glDisable( m_nGLTextureType );

	        Gl.glPushAttrib( Gl.GL_TRANSFORM_BIT );
	        Gl.glMatrixMode( Gl.GL_TEXTURE );
	        Gl.glPopMatrix();
	        Gl.glPopAttrib();
        }


        public void FreeTexture() {
	        if ( m_nGLTextureID != 0 ) {
                if ( m_nDeleteGLTextureID != 0 )
                    System.Diagnostics.Debugger.Break();    // leaked texture ID!
		        m_nDeleteGLTextureID = m_nGLTextureID;
	        }
	        m_nGLTextureID = 0;
        }



        void OnImageChanged() {
            FreeTexture();

            int nWidth = m_image.Width;
            int nHeight = m_image.Height;

            //  figure out scaling factors
            m_nTexWidth = 1;
            while (m_nTexWidth < nWidth)
                m_nTexWidth *= 2;
            float fScaleX = (float)m_nTexWidth / (float)nWidth;
            m_nTexHeight = 1;
            while (m_nTexHeight < nHeight)
                m_nTexHeight *= 2;
            float fScaleY = (float)m_nTexHeight / (float)nHeight;

            m_fScaleX = 1.0f / fScaleX;
            m_fScaleY = 1.0f / fScaleY;

            //m_fScaleX = (fScaleX < fScaleY) ? fScaleX : fScaleY;
            //m_fScaleX = 1.0f / m_fScaleX;	// scale tex coords *down* to scale image *up*

            m_nGLTextureFormat = Gl.GL_RGBA;
        }

        int GetTextureID() {
            if ( m_nGLTextureID != 0 )
                return m_nGLTextureID;

            // get rid of old texture ID now that we are in GL context
            if ( m_nDeleteGLTextureID != 0 ) {
                Gl.glDeleteTextures(1, ref m_nDeleteGLTextureID);
                m_nDeleteGLTextureID = 0;
            }

            Gl.glGenTextures(1, out m_nGLTextureID);

            Gl.glEnable(m_nGLTextureType);
            Gl.glBindTexture(m_nGLTextureType, m_nGLTextureID);

            Gl.glTexParameteri(m_nGLTextureType, Gl.GL_TEXTURE_MIN_FILTER, m_nGLTextureFilter);
            Gl.glTexParameteri(m_nGLTextureType, Gl.GL_TEXTURE_MAG_FILTER, m_nGLTextureFilter);

            Gl.glTexParameteri(m_nGLTextureType, Gl.GL_TEXTURE_WRAP_S, m_nGLTextureWrapMode);
            Gl.glTexParameteri(m_nGLTextureType, Gl.GL_TEXTURE_WRAP_T, m_nGLTextureWrapMode);

            int nBorder = 0;
            Gl.glTexImage2D(m_nGLTextureType, 0, m_nGLTextureFormat, 
                         m_nTexWidth + 2 * nBorder, m_nTexHeight + 2 * nBorder, nBorder,
                          m_nGLTextureFormat, Gl.GL_UNSIGNED_BYTE, (string)null);

            int xoff = (m_nTexWidth - m_image.Width) / 2;
            int yoff = (m_nTexHeight - m_image.Height) / 2;
            Gl.glTexSubImage2D(m_nGLTextureType, 0, xoff, yoff, 
                m_image.Width, m_image.Height, m_nGLTextureFormat, Gl.GL_UNSIGNED_BYTE, m_image.Buffer );

            Gl.glBindTexture(m_nGLTextureType, 0);
            Gl.glDisable(m_nGLTextureType);

            return m_nGLTextureID;
        }
    }
}
