using System;
using System.Collections.Generic;
using System.Text;

namespace Sketcher
{
    class IconSet
    {
        //! default states (user can also define their own states)
        public static readonly int IconState_Enabled = 0;
        public static readonly int IconState_Disabled = 1;
        public static readonly int IconState_Hilighted = 2;

        public static readonly int Resolution_Max = int.MaxValue;

        public static readonly Icon DefaultIcon = MakeDefaultIcon();

        //! icons are indexed first by state identifier, then by resolution
        SortedList<int,
            SortedList<int, Icon>> m_icons;

        public IconSet( ) {
            m_icons = new SortedList<int, SortedList<int, Icon>>();
        }

        public void SetIcon( int nState, int nResolution, Icon icon ) {
            if ( ! m_icons.ContainsKey(nState) )
                m_icons.Add(nState, new SortedList<int,Icon>() );
            m_icons[nState].Add(nResolution, icon);
        }

        public void SetIcon( int nState, Icon icon ) {
            int nResolution = icon.IconImage.Width;
            SetIcon( nState, nResolution, icon );
        }


        public Icon GetBestIcon( int nState ) {
            return GetBestIcon( nState, Resolution_Max );
        }
        public Icon GetBestIcon( int nState, Box2f bounds ) {
            return GetBestIcon( nState, (int)Math.Ceiling(bounds.Width) );
        }


        public Icon GetBestIcon( int nState, int nResolution ) {

            // find icons for this state, or return default icon
            SortedList<int, Icon> stateIcons = null;
            if ( m_icons.ContainsKey(nState) )
                stateIcons = m_icons[nState];
            else if ( m_icons.ContainsKey(IconState_Enabled) )
                stateIcons = m_icons[IconState_Enabled];
            else if (m_icons.Values.Count > 0) {
                stateIcons = m_icons.Values[0];
            } else
                stateIcons = null;
            if ( stateIcons == null )
                return DefaultIcon;
            
            // find best resolution to use for this state
            if ( nResolution == Resolution_Max )
                return stateIcons.Values[ stateIcons.Values.Count-1 ];
            else if (stateIcons.ContainsKey( nResolution ))
                return stateIcons[nResolution];
            else {
                for (int i = 0; i < stateIcons.Keys.Count; ++i) {
                    if (stateIcons.Keys[i] >= nResolution)
                        return stateIcons[stateIcons.Keys[i]];
                }
                return stateIcons[stateIcons.Keys[stateIcons.Keys.Count-1]];
            }
        }


        public void LoadBasicIconSet( FilePath sEnabledFile, FilePath sDisabledFile, FilePath sHilightedFile ) {
            if (sEnabledFile != null) {
                Icon icon = Icon.LoadIconFile( sEnabledFile );
                if (icon != null)
                    SetIcon( IconState_Enabled, icon );
            }

            if (sDisabledFile != null) {
                Icon icon = Icon.LoadIconFile( sDisabledFile );
                if (icon != null)
                    SetIcon( IconState_Disabled, icon );
            }

            if (sHilightedFile != null) {
                Icon icon = Icon.LoadIconFile( sHilightedFile );
                if (icon != null)
                    SetIcon( IconState_Hilighted, icon );
            }
        }



        static Icon MakeDefaultIcon( ) {
            Image defaultImg = new Image( 2, 2, 4 );
            defaultImg.SetPixel( 0, 0, new Image.Pixel( ColorRGBA.RED ) );
            defaultImg.SetPixel( 0, 1, new Image.Pixel( ColorRGBA.GREEN ) );
            defaultImg.SetPixel( 1, 0, new Image.Pixel( ColorRGBA.GREEN ) );
            defaultImg.SetPixel( 1, 1, new Image.Pixel( ColorRGBA.BLUE ) );
            Icon defaultIcon = new Icon( defaultImg );
            defaultIcon.ScaleMode = Icon.ScaleModeType.Nearest;
            return defaultIcon;
        }

    }
}
