using System;
using System.Collections.Generic;
using System.Text;


namespace Sketcher
{
    class Stroke
    {
        List<Vector2f> m_vPosition;
        List<double> m_vTime;
        List<float> m_vPressure;

        bool m_bStrokeStatsValid;

        List<float> m_vAngle;
        Box2f m_bounds;
        float m_fTotalLength;
        float m_fAngleSum;
        float m_fWeightedWindingAngle;
        float m_fWeightSum;

        SysTimer s;

        public Stroke() {
            m_vPosition = new List<Vector2f>();
            m_vTime = new List<double>();
            m_vPressure = new List<float>();

            // stroke stats get computed and cached on demand
            m_vAngle = null;
            m_bounds = new Box2f();
            m_fAngleSum = 0;
            m_fTotalLength = 0;
            m_fWeightedWindingAngle = 0;
            m_fWeightSum = 0;
            m_bStrokeStatsValid = false;

            s = new SysTimer();
            s.Start();
        }


        public int Count {
            get { return m_vPosition.Count; }
        }


        public List<Vector2f> Vertices {
            get { return m_vPosition; }
        }
        public List<double> Times {
            get { return m_vTime; }
        }
        public List<float> Pressures {
            get { return m_vPressure; }
        }

        public void DoPressureHack(float fStartPressure) {
            m_vPressure[0] = fStartPressure;
        }


        public Vector2f this[int i] {
            get { return m_vPosition[i]; }
        }
        public Vector2f First {
            get { return m_vPosition[0]; }
        }
        public Vector2f Last {
            get { return m_vPosition[m_vPosition.Count - 1]; }
        }
        public Vector2f LastDelta {
            get { return m_vPosition[m_vPosition.Count - 1] - m_vPosition[m_vPosition.Count - 2]; }
        }


        public Vector2f EndsDirection {
            get { return ( m_vPosition.Count == 0 ) ? Vector2f.ZERO : 
                    ( m_vPosition[m_vPosition.Count - 1] - m_vPosition[0] ).Normalized();           }
        }
        public float EndsDistance {
            get { return ( m_vPosition.Count == 0 ) ? 0 : 
                    ( m_vPosition[m_vPosition.Count - 1] - m_vPosition[0] ).Length;           }
        }


        public List<float> Angles {
            get { ValidateStrokeStats(); return m_vPressure; }
        }
        public Box2f Bounds {
            get { ValidateStrokeStats(); return m_bounds; }
        }
        public float TotalLength {
            get { ValidateStrokeStats();  return m_fTotalLength; }
        }
        public float WindingAngle {
            get { ValidateStrokeStats(); return (float)Math.Abs(m_fAngleSum) * (180.0f / (float)Math.PI); }
        }
        public float SignedWindingAngle {
            get { ValidateStrokeStats(); return m_fAngleSum * (180.0f / (float)Math.PI); }
        }
        public float AverageWindingAngle {
            get { ValidateStrokeStats(); return (float)Math.Abs(m_fAngleSum / (float)m_vPosition.Count) * (180.0f / (float)Math.PI); }
        }
        public float WeightedAvgWindingAngle {
            get { ValidateStrokeStats(); return (m_fWeightedWindingAngle / m_fWeightSum) * (180.0f / (float)Math.PI); }
        }

        public bool AddVertex(float fX, float fY, float fPressure) {
            //if ( Count > 0 && m_vPosition[m_vPosition.Count-1].X == fX && m_vPosition[m_vPosition.Count-1].Y == fY )
            //    return false;
            if (Count > 0 && m_vPosition[m_vPosition.Count-1].SquaredDist( fX, fY ) < 1.0f)
                return false;

            Vector2f vNew = new Vector2f(fX, fY);
            m_vPosition.Add(vNew);
            m_vTime.Add(s.Current);
            m_vPressure.Add(fPressure);

            m_bStrokeStatsValid = false;

            return true;
        }


        void ValidateStrokeStats() {
            if (m_bStrokeStatsValid)
                return;
            if ( m_vPosition.Count == 0 )
                return;

            m_fAngleSum = 0;
            m_fTotalLength = 0;
            m_fWeightedWindingAngle = 0;
            m_fWeightSum = 0; 
            m_bounds = new Box2f(m_vPosition[0], m_vPosition[0]);
            m_vAngle = new List<float>(m_vPosition.Count);
            m_vAngle.Add(0);

            for ( int i = 1; i < m_vPosition.Count-1; ++i ) {
                m_bounds.Contain(m_vPosition[i]);

                int nMiddle = i;
                int nBefore = nMiddle - 1;
                int nAfter = nMiddle + 1;
                Vector2f dx1 = m_vPosition[nMiddle] - m_vPosition[nBefore];
                Vector2f dx2 = m_vPosition[nAfter] - m_vPosition[nMiddle];
                    
                float d1 = dx1.Normalize(); 
                float d2 = dx2.Normalize();
                m_fTotalLength += d1;

                float fCross = dx1.Cross(dx2);
                float fSign = (fCross > 0) ? 1 : -1;

                m_vAngle.Add(fSign * (float)Math.Acos(MathUtil.Clamp(dx1.Dot(dx2), -1, 1)));

                m_fAngleSum += m_vAngle[nMiddle];
                if (float.IsNaN(m_fAngleSum))
                    System.Diagnostics.Debugger.Break();

                float fWeight = 0.5f * (d1 + d2);
                m_fWeightedWindingAngle += Math.Abs(m_vAngle[nMiddle]) * fWeight;
                m_fWeightSum += fWeight;
            }

            m_vAngle.Add(0);

            m_bStrokeStatsValid = true;
        }



        public float IsLineStroke() {

            // find line
            Vector2f vStart = this[0];
            Vector2f vDir = this[Count - 1] - vStart;
            vDir.Normalize();

            // find total deviation from line, measured as 1 - (angle between segment and line)
            float fDeviation = 0.0f;
            for (int i = 1; i < m_vPosition.Count; ++i) {
                Vector2f vSeg = m_vPosition[i] - m_vPosition[i - 1];
                vSeg.Normalize();
                fDeviation += 1 - Math.Abs(vDir.Dot(vSeg));
            }
            float fAvgDev = fDeviation / (float)(Count - 1);

            const float fDevThresh = 0.25f;
            float fFactor = MathUtil.WyvillFalloff01( fAvgDev / fDevThresh );

            return fFactor;
        }

        public float IsCircleStroke() {
            const float fMinWindingAngle = 180.0f;
            if (WindingAngle < fMinWindingAngle)
                return 0;

            float fAngleFactor = MathUtil.SmoothRise0To1(WindingAngle, 0.05f, fMinWindingAngle, 0.42f);
            float fBoundsFactor = Math.Min(m_bounds.Width, m_bounds.Height) / Math.Max(m_bounds.Width, m_bounds.Height);

            return fAngleFactor * fBoundsFactor;
        }


        //! stroke is click if distance total stroke size is < constant
        public float IsClickStroke() {
            const float fMaxSize = 3.0f;

            return Math.Max(1 - (m_bounds.DiagonalLen / fMaxSize), 0);
        }








        //! assumes pressure is in range [0,1]
        public static float PressureToSizeMap(float fPressure, float fMinSize, float fMaxSize) { 
            float fT = 1-fPressure*fPressure;
            fT = fT*fT*fT;
            fT = 1-fT;
            return (1-fT)*fMinSize + (fT)*fMaxSize;
        }

        public static float PressureToOpacityMap(float fPressure, float fMinOpacity, float fMaxOpacity) {
            float fT = 1-fPressure*fPressure;
            fT = fT*fT*fT;
            fT = 1-fT;
            return (1-fT)*fMinOpacity + (fT)*fMaxOpacity;
        }



    }
}
